"""
Utilities for various components. Aims to contain any resuable functions needed elsewhere
"""
import os
import pathlib
import time
from pathlib import Path

import numpy as np
import xarray as xr
import zarr.errors
from numba import vectorize, float64
from scipy.interpolate import interp1d


@vectorize([float64(float64, float64, float64, float64)])
def inv_sin_sqr(y, mag, x_offset, y_offset):
    """Function to transform calibration. Superseded by interpolation
    :param y:
    :param mag:
    :param x_offset:
    :param y_offset:
    :return:
    """
    return np.mod(((180 / np.pi) * (np.arcsin(np.sqrt(np.abs((y - y_offset) / mag)))) + x_offset), 180)


@vectorize([float64(float64, float64, float64, float64)])
def sin_squared(x, mag, x_offset, y_offset):
    """
    Function that squares the sin of a number
    :param y:
    :param mag:
    :param x_offset:
    :param y_offset:
    :return:
    """
    return ((np.sin((x - x_offset) * np.pi / 180)) ** 2) * mag + y_offset


def interp(old, pol, pwr) -> np.array:
    """
    Interpolates a set of powers for polarization values to generate a new set of polarizations for given power
    :param old: original power values
    :param pol: polarization values
    :param pwr: new power values
    :return: list of polarizations for the new power values
    """
    function = interp1d(old, pol, fill_value="extrapolate")
    return function(pwr)


def interpolate(filename: pathlib.PosixPath, pwr: np.array = np.arange(0, 100, 5), throw: int = 0) -> xr.DataArray:
    """
    Interpolates a calibration file to get Polarizations for given powers and wavelength
    :param throw: throw everything at this polarization and lower
    :type throw: int
    :param filename: calibration file to interpolate from
    :type filename: pathlib.PosixPath
    :param pwr: powers to interpolate to
    :type pwr: np.array
    :return: Reversed calibration
    :rtype: xr.DataArray
    """
    power_calibration = xr.open_dataset(filename, engine="zarr")["Pwr"]
    power_calibration = power_calibration.where(power_calibration.Polarization > throw, drop=True)
    pc_pol = power_calibration.coords["Polarization"].values
    pc_reverse = xr.apply_ufunc(interp, power_calibration, input_core_dims=[["Polarization"]], vectorize=True,
                                output_core_dims=[["power"]], kwargs={"pwr": pwr, "pol": pc_pol})
    pc_reverse.coords["power"] = pwr
    return pc_reverse


def scan_directory(extensions: dict) -> dict:
    """
    Scans for readable files

    :return:
    :param extensions: A mapping from data types to modules
    :type extensions: dict
    :return: Mapping from files to modules
    :rtype: dict
    """
    # parses directories for valid files
    files = list(Path("").rglob("*.zarr"))
    file_dict = {}
    for file in files:
        try:
            dataset = xr.open_dataset(file, engine="zarr")
            data_type = dataset.attrs["data_type"]
            dataset.close()
            if data_type in extensions:
                file_dict[file] = extensions[data_type]
        except zarr.errors.GroupNotFoundError:
            print(f"{file} open failed. No data in folder")
    return file_dict


def extension(file: str) -> os.path:
    """

    :param file:
    :type file: string or path
    :return:
    :rtype: path
    """
    return os.path.splitext(file)[1][1:]


def fname(file) -> str:
    """
    Helper function to get the name from a filename
    (IE: data/truncated_1.zarr -> truncated_1)

    :param file: filename or path
    :type file: str
    :return: just the filename with no extension
    :rtype: str
    """
    file_name = os.path.basename(file)
    return file_name.replace(extension(file_name), '')


def hotfix(dataset):
    """
    Fixes older datasets by adding coordinates and atttributes. DO NOT USE
    :param dataset:
    :type dataset: xr.Dataset
    :return: fixed dataset
    :rtype: xr.Dataset
    """
    dataset.coords['wavelength'] = dataset.coords['wavelength'] * 2 + 780
    dataset.coords['Polarization'] = np.arange(0, 180, 1) / 90 * np.pi
    dataset.coords['degrees'] = ("Polarization", np.arange(0, 360, 2))
    dataset.coords['x_pixels'] = ("x", dataset.coords['x'].values)
    dataset.coords['y_pixels'] = ("y", dataset.coords['y'].values)
    dataset.coords['x'] = dataset.coords['x'] * 0.05338  # TODO fix incorrect magic numbers
    dataset.coords['y'] = dataset.coords['y'] * 0.05338  # TODO fix incorrect magic numbers
    dataset.attrs['x'] = "micrometers"
    dataset.attrs['y'] = "micrometers"
    dataset.attrs['Polarization'] = "radians"
    return dataset


def get_range(dim: str, coords: xr.core.coordinates.DataArrayCoordinates) -> tuple[np.float64, np.float64]:
    """
    Gets maximum and minimum on a set of coordinates

    :param dim: dimension to get range on
    :type dim: str
    :param coords: coordinates
    :type coords: xarray.core.coordinates.DataArrayCoordinates
    :return: minimum and maximum values on dimension
    :rtype: Tuple[numpy.float64, numpy.float64]
    """
    return coords[dim].values.min(), coords[dim].values.max()


def convert(seconds):
    """

    :param seconds:
    :type seconds:
    :return:
    :rtype:
    """
    return time.strftime("%H:%M:%S", time.gmtime(seconds))



import holoviews as hv
import simple_pid
import param
import panel as pn
from . import ensemble_RASHG

name = "RASHG_PID"

hv.extension('bokeh')
raise Exception("This code doesn't even work why are you using it???")


class Ensemble(ensemble_RASHG.Ensemble):
    pid_time = param.Number(default=1)
    pid = simple_pid.PID()

    def __init__(self):
        super().__init__()

    def pid_step(self):
        control = self.pid(value)
        self.atten.move_abs(control)

    def initialize(self):
        super().initialize()
        self.pid.sample_time = self.pid_time
        pn.state.add_periodic_callback(self.pid_step, self.pid_time)

    def widgets(self):
        if self.initialized:
            return pn.Column(super().widgets)
        else:
            return None



"""
Specifies base class
"""
from array import array
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Union, ClassVar

import numpy as np
import param


@dataclass
class Coordinate:
    name: str
    unit: str
    dimension: str
    values: np.array = np.array([])
    step_function: Union[Callable[[array], None], str] = field(default="none")

    def __call__(self, coords: array) -> None:
        if self.step_function != "none":
            self.step_function(coords)


@dataclass
class Coordinates:
    """Essentially a hybrid between a dictionary and a list"""
    coordinates: list[Coordinate]
    lookup_table: ClassVar[dict]

    def __post_init__(self):
        self._gen_lookup_table()

    def append(self, coordinate: Coordinate):
        self.coordinates.append(coordinate)
        self._gen_lookup_table()

    def _gen_lookup_table(self):
        self.lookup_table = {coordinate.name: i for i, coordinate in enumerate(self.coordinates)}

    def __iter__(self):
        yield from self.coordinates

    def __getitem__(self, indices: str) -> Coordinate:
        return self.coordinates[self.lookup_table[indices]]


class EnsembleBase(param.Parameterized):
    """
    Base class for all ensembles. Other instrument groups should inherit from this
    """
    initialized: bool = param.Boolean(default=False, precedence=-1)  # dummy variable to make code work
    type: str = "base"
    title: str = param.String(default="Power/Wavelength dependent RASHG")
    filename: str = param.String(default="data/testfolder.zarr")
    datasets: array = ["ds1"]
    live: bool = True
    gather: bool = True
    coords: Coordinates

    def initialize(self):
        """
        Initializes the instrument. Runs when you hit confirm
        :return:
        """
        pass

    def get_frame(self, coords):
        """
        Captures the frame
        :param coords: list of data in order of loop_coords
        :return: data
        """
        pass

    def widgets(self):
        """
        Widgets to pass to dashboard
        :return:
        """
        return self.param

    def start(self):
        """
        Any functions to run at start
        :return:
        """
        pass

    def stop(self):
        """
        Any functions to run at stop
        :return:
        """
        pass

    def graph(self, live=False):
        """
        returns a graph
        :param live: whether or not it is a live view
        :type live: bool
        :return:
        """
        return None



"""
Various Utilities for command line. Formerly converter.py
"""
import os
from pathlib import Path
import zarr
import xarray as xr
from zarr.errors import GroupNotFoundError
from neogidashboard import utils

compressor = zarr.Blosc(cname="zstd", clevel=3, shuffle=2)


def convert():
    """Converts all avalible files to zarr files and fills in missing attributes. Only to be used to update legacy
    files. """
    # client = Client()
    for file in list(Path("..").rglob("*.zarr")):
        print(file)
        ds1 = xr.open_dataset(file, engine="zarr")
        if not "data_type" in ds1.attrs:
            ds1.attrs["data_type"] = "RASHG"
            ds1.to_zarr(file, mode="w", compute=True)
    for file in list(Path("..").rglob("*.5nc")):
        filename = str(file).replace(f".{utils.extension(file)}", '.zarr')
        if filename not in list(Path("..").rglob("*.zarr")):
            dataset = utils.hotfix(xr.open_dataarray(file, engine="netcdf4"))
            coords = dataset.coords
            ds_coords = dataset.assign_coords(power=0).expand_dims("power")
            data = xr.Dataset(data_vars={"ds1": ds_coords},
                              attrs=dataset.attrs,
                              coords=coords)
            data.attrs["data_type"] = "RASHG"
            data.attrs["title"] = "RASHG"
            data.attrs["data_version"] = 2
            print(data)
            data.to_zarr(filename, encoding={"ds1": {"compressor": compressor}}, consolidated=True, compute=True)


def clean():
    """cleans all your empty files"""
    files = list(Path("").rglob("*.zarr"))
    for file in files:
        try:
            dataset = xr.open_dataset(file, engine="zarr")
            dataset.close()
        except GroupNotFoundError:
            os.removedirs(file)


if __name__ == '__main__':
    convert()



"""
Module for viewing calibration files
"""
from pathlib import Path, PosixPath

import holoviews as hv
import numpy as np
import panel as pn
import param
import xarray as xr

from ... import utils

pn.extension('plotly')
hv.extension('bokeh')

DATA_TYPE = "WavelengthPoweredCalib"


class Grapher(param.Parameterized):
    """visualizer for calibration files"""
    dataset = param.ObjectSelector(objects=["Pwr", "Pwrstd", "Vol", "Volstd"], default="Pwr")
    wavelength = param.Selector()
    pstart = param.Integer(default=0)
    pstop = param.Integer(default=100)
    pstep = param.Number(default=5)
    throw = param.Number(default=0)

    def _update_dataset(self):
        """
        Opens selected dataset
        """
        self.interpolate()
        self.data = xr.open_zarr(self.filename)
        self.param['wavelength'].objects = self.data["Pwr"].coords['wavelength'].values.tolist()
        self.wavelength = self.data["Pwr"].coords["wavelength"].min().values

    @param.depends("pstart", "pstop", "pstep","throw", watch=True)
    def interpolate(self):
        power = np.arange(self.pstart, self.pstop, self.pstep)
        self.pc_reverse = utils.interpolate(PosixPath(self.filename), pwr=power,throw=self.throw)
        self.pol_dim = hv.Dimension('Polarization', soft_range=(0, 10), unit="degrees")
        self.pwr_dim = hv.Dimension("Power", range=(self.pstart, self.pstop))

    def __init__(self, filename, client_input):
        """
        :param filename: filename to open
        :param client_input: Dask Client
        """
        super().__init__()
        self.client = client_input
        self.filename = Path(filename)
        self._update_dataset()

    @param.depends("wavelength", "dataset")
    def nav(self):
        """
        Navigation graph
        :return:
        """
        output = self.data[self.dataset].sel(wavelength=self.wavelength)
        opts = [hv.opts.Curve(title=f"Wavelength: {self.wavelength}",
                              tools=['hover'], framewise=True)]
        return hv.Curve(output).opts(opts)

    @param.depends("wavelength", "pstart", "pstop", "pstep")
    def reverse_nav(self):
        """Reversed navigation graph"""
        output = self.pc_reverse.sel(wavelength=self.wavelength)
        opts = [hv.opts.Curve(title=f"Wavelength: {self.wavelength}",
                              tools=['hover'], framewise=True)]
        return hv.Curve(output, "power", "polarization").opts(opts).redim(polarization=self.pol_dim, power=self.pwr_dim)

    def view(self):
        """
        wraps nav
        :return: navigation
        """
        return pn.Row(self.nav, self.reverse_nav)

    def widgets(self):
        """
        returns widgets
        :return: widgets
        """
        widgets = {"wavelength": pn.widgets.DiscreteSlider}
        return pn.Param(self.param, widgets=widgets)



"""
Wraps visualizer and ensembles to serve dashboard. Also includes command line parsing
"""
from typing import Union, Optional
import argparse
import time
import socket
import sys
import holoviews as hv
import param
import panel as pn
from .visualizer import types, Viewer
from .ensembles import instruments, gui

pn.extension('plotly')
hv.extension('bokeh', 'plotly')



def is_port_in_use(port):
    """
    Checks if port is in use

    :param port: port to check
    :type port: int
    :return: if port is in use
    :rtype: bool
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as socket_to_check:
        return socket_to_check.connect_ex(('localhost', port)) == 0


# wrapper around viewer class to interface with instrumental class
class Combined(param.Parameterized):
    """
    Combined viewer for Visualizer and Instruments
    """
    applet: Union[gui.Gui, Viewer]
    applets = ["viewer", "instrumental"]
    applets = param.ObjectSelector(default="instrumental", objects=applets)
    button = pn.widgets.Button(name="STOP", button_type='primary')

    def __init__(self) -> None:
        """

        :rtype: None
        """
        super().__init__()
        self.applet = gui.Gui(instruments)

    @param.depends('applets', watch=True)
    def load(self) -> None:
        """
        Loads the currently selected applet

        :rtype: None
        """
        if self.applets == "viewer":
            self.applet = Viewer(types)
        elif self.applets == "instrumental":

            self.applet = gui.Gui(instruments)

    @param.depends('applets')
    def widgets(self) -> Union[pn.Row, pn.Column]:
        """
        Renders widgets from the applet

        :return: widgets
        :rtype: pn.Column
        """
        return self.applet.widgets

    @param.depends('applets')
    def applet_view(self) -> Union[Optional[pn.Row], pn.Column]:
        """
        wraps applet view

        :return: applet view
        :rtype: pn.Column
        """
        return self.applet.graph

    def quit(self, event: tuple = None) -> None:
        """
        closes dashboard

        :param event: allows this to be used with a button
        :type event: tuple
        """
        self.applet.stop()
        sys.exit()

    def view(self) -> pn.layout.base.Row:
        """
        Returns widgets and applet view


        :return: dashboard to be rendered
        :rtype: panel.layout.base.Row
        """
        self.button.on_click(self.quit)
        return pn.Row(pn.Column(self.param, self.widgets, self.button), self.applet_view)


# these two functions are basically identical for now
def serve(port: int = 5006, open_browser: bool = True) -> None:
    """
    serves the combined panel applet

    Parameters
    ----------
    :rtype: None
    :param open_browser: whether to open the browser
    :type open_browser: bool
    :param port: default port number to use
    :type port: int
    """
    view = Combined()
    while is_port_in_use(port):
        port += 1
    view.view().show(port=port, open=open_browser)
    # if you need to change this, change this on your own or implement ports yourself. It
        # isn't very hard



def main() -> None:
    """

    :rtype: None
    """
    parser = argparse.ArgumentParser(prog='combined', description='Deploys and runs panel server')
    parser.add_argument('-server', dest='server', help='Runs the panel server for multiple clients',
                        action='store_const', const=True, default=False)
    parser.add_argument('-local', dest='local', help='Runs the panel server for single clients', action='store_const',
                        const=True, default=False)
    parser.add_argument('--fit', dest='filename', help='fits datafile and saves to file from command line',
                        action='store', default=False)
    args = parser.parse_args()
    if args.server:
        serve(open_browser=False)
    elif args.local:
        serve()
    elif args.filename:
        start = time.time()
        view = Viewer(filename=args.filename, types=types)  # use port 8787 to view stats
        view.view().show()
        end = time.time()
        print(end - start)
    else:
        print("Defaulting to local server")
        serve()


if __name__ == '__main__':
    main()



import time
from array import array

import neogiinstruments
import numpy as np
import panel as pn
import param

from ..ensemblebase import EnsembleBase, Coordinate, Coordinates
from ... import utils

name = "stellarnet"


def get_calibs() -> list:
    """
    Scans for calibration files

    :return: list of all calibration files
    :rtype: list of PosixPath
    """
    return list(utils.scan_directory({"WavelengthPoweredCalib": None}).keys())


class Ensemble(EnsembleBase):
    wavstart = param.Integer(default=780)
    wavend = param.Integer(default=900)
    wavstep = param.Integer(default=2)
    pstart = param.Integer(default=0)
    pstop = param.Integer(default=10)
    pstep = param.Number(default=0.5)
    mai_time = param.Integer(default=30)
    pwait = param.Integer(default=1)
    type = name
    data = "stellarnet"
    datasets = ["Stellarnet", "V", "Vstd"]
    dimensions = {"Stellarnet": ["wavelength", "power", "emission_wavelength"], "V": ["wavelength", "power"],
                  "Vstd": ["wavelength", "power"]}
    cap_coords = {"Stellarnet": ["emission_wavelength"], "V": [], "Vstd": []}
    loop_coords = ["wavelength", "power"]
    debug = param.Boolean(default=False)
    live = False
    calibration_file = param.ObjectSelector()

    def __init__(self):
        files = get_calibs()
        if len(files) == 0:
            print("Needs calibration file ")
        self.param["calibration_file"].objects = files
        self.param["calibration_file"].default = files[0]
        super().__init__()
        self.filename = "data/stellarnet.zarr"
        self.rotator = neogiinstruments.rotator("rotator")
        self.MaiTai = neogiinstruments.MaiTai()
        self.StellarNet = neogiinstruments.StellarNet()
        self.Photodiode = neogiinstruments.Photodiode()
        self.coords = Coordinates([
            Coordinate("wavelength", "nanometer", "wavelength", step_function=self.wav_step),
            Coordinate("power", "degrees", "power", step_function=self.pow_step),
            Coordinate("emission_wavelength", "nanometers", "emission_wavelength")]
        )

    def wav_step(self, xs):
        self.MaiTai.instrument.Set_Wavelength(xs[0])
        if self.debug:
            print(f'moving to {xs[0]}')
        time.sleep(self.mai_time)
        self.pow_step(xs)
        if not self.debug:
            time.sleep(10)
        self.MaiTai.instrument.Shutter(1)
        if self.debug:
            print(f'starting loop at {xs[0]}')
        if not self.debug:
            time.sleep(5)

    def initialize(self):
        self.initialized = True
        exclude = []
        for param in self.param:
            if not param in exclude:
                self.param[param].constant = True

        self.init_vars()

    def start(self):
        self.rotator.instrument.home()

    def init_vars(self):
        self.coords["wavelength"].values = np.arange(self.wavstart, self.wavend, self.wavstep, dtype=np.uint16)
        power = np.arange(self.pstart, self.pstop, self.pstep)
        self.coords["power"].values = power
        emission_wavelength = self.StellarNet.instrument.GetSpec()[0]
        self.emission_length = len(emission_wavelength)
        self.coords["emission_wavelength"].values = emission_wavelength
        self.pc_reverse = utils.interpolate(self.calibration_file, pwr=power)

    def pow_step(self, xs: array):
        pow = xs[1]
        wav = xs[0]
        pol = self.pc_reverse.sel(power=pow, wavelength=wav).values
        if self.debug:
            print(f"moving to {pol}")
        if -360 < pol < 360:
            self.rotator.instrument.move_abs(pol)
        time.sleep(self.pwait)

    def get_frame(self, coords):
        data = self.StellarNet.instrument.GetSpec()[1]
        V, Vstd = self.Photodiode.instrument.gather_data()
        return {"Stellarnet": data, "V": V, "Vstd": Vstd}

    def widgets(self):
        if self.initialized:
            return pn.Column(self.rotator.view, self.StellarNet.view, self.MaiTai.view, self.Photodiode.view)
        else:
            return None



import time

import neogiinstruments
import numpy as np
import panel as pn
import param

from ..ensemblebase import EnsembleBase, Coordinate, Coordinates

name = "WavelengthPoweredCalib"


class Ensemble(EnsembleBase):
    wavstart = param.Integer(default=780)
    wavend = param.Integer(default=800)
    wavstep = param.Integer(default=2)
    pstart = param.Integer(default=0)
    pstop = param.Integer(default=10)
    pstep = param.Number(default=0.5)
    pwait = param.Integer(default=1)
    mai_time = param.Integer(default=30)
    type = name
    data = "WavelengthPoweredCalib"
    dimensions = ["wavelength", "Polarization"]
    cap_coords = []
    loop_coords = ["wavelength", "Polarization"]
    datasets = ["Pwr", "Pwrstd", "Vol", "Volstd"]
    debug = param.Boolean(default=False)
    live = False
    def start(self):
        self.rotator.home()
    def __init__(self):
        super().__init__()
        self.filename = "calib/WavelengthPowerCalib.zarr"
        self.rotator = neogiinstruments.rotator("rotator")
        self.MaiTai = neogiinstruments.MaiTai()
        self.PowerMeter = neogiinstruments.PowerMeter()
        self.Photodiode = neogiinstruments.Photodiode()

    def wav_step(self, xs):
        if self.debug:
            print(f'moving to {xs[0]}')
        self.MaiTai.instrument.Set_Wavelength(xs[0])
        time.sleep(self.mai_time)
        self.pol_step(xs)
        if not self.debug:
            time.sleep(10)
        self.MaiTai.instrument.Shutter(1)
        if self.debug:
            print(f'starting loop at {xs[0]}')
        if not self.debug:
            time.sleep(5)

    def initialize(self):
        self.initialized = True
        exclude = []
        for parameter in self.param:
            if parameter not in exclude:
                self.param[parameter].constant = True

        self.init_vars()
        self.coords = Coordinates(
            [
                Coordinate("wavelength", "nanometer", "wavelength", self.wavelength, self.wav_step),
                Coordinate("Polarization", "degrees", "Polarization", self.Polarization, self.pol_step)
            ]
        )

    def init_vars(self):
        self.wavelength = np.arange(self.wavstart, self.wavend, self.wavstep, dtype=np.uint16)
        self.Polarization = np.arange(self.pstart, self.pstop, self.pstep)

    def pol_step(self, xs):
        pol = xs[1]
        if self.debug:
            print(f"moving to {pol}")
        self.rotator.instrument.move_abs(pol)
        time.sleep(self.pwait)

    def get_frame(self, coords):
        if self.debug:
            print("Gathering power data")
        p = self.PowerMeter.instrument.PowAvg()
        Pwr = p[0]
        Pwrstd = p[1]
        if self.debug:
            print("Gathering Photodiode data")
        V, Vstd = self.Photodiode.instrument.gather_data()
        if self.debug:
            print(f"Pwr: {Pwr}, Pwrstd: {Pwrstd}, Vol: {V}, Volstd: {Vstd}")
        return {"Pwr": Pwr, "Pwrstd": Pwrstd, "Vol": V, "Volstd": Vstd}

    def widgets(self):
        if self.initialized:
            return pn.Column(self.rotator.view, self.PowerMeter.view, self.Photodiode.view, self.MaiTai.view)
        else:
            return None



"""
Visualizer for RASHG data
"""
import time
import os
from pathlib import Path
from datetime import timedelta
import numpy as np
from numba import vectorize, float64
import pandas as pd
import param
import plotly.express as px
import xarray as xr
import holoviews as hv
import panel as pn
from holoviews import streams
from ... import utils

pn.extension('plotly')
hv.extension('bokeh')

DATA_TYPE = "RASHG"


@vectorize([float64(float64, float64, float64, float64, float64, float64)])
def function(phi: float64, delta: float64, A: float64, B: float64, theta: float64, C: float64) -> float64:
    """
    Function to fit to, accelerated using numba
    :param phi:
    :type phi:
    :param delta:
    :type delta:
    :param A:
    :type A:
    :param B:
    :type B:
    :param theta:
    :type theta:
    :param C:
    :type C:
    :return:
    :rtype: float64

    """
    return (A * np.cos(3 * phi - 3 * delta) + B * np.cos(phi - 3 * delta + 2 * theta)) ** 2 + C


class Grapher(param.Parameterized):
    """
    Visualizer for RASHG data
    """
    zdim: hv.Dimension
    orientation = param.Integer(default=0, bounds=(0, 1))
    wavelength = param.Selector()
    power = param.Selector()
    x0 = param.Number(default=0, precedence=-1)
    x1 = param.Number(default=1, precedence=-1)
    y0 = param.Number(default=0, precedence=-1)
    y1 = param.Number(default=1, precedence=-1)
    selected = param.Boolean(default=False, precedence=-1)
    colorMap = param.ObjectSelector(default="fire", objects=hv.plotting.list_cmaps())
    button = pn.widgets.Button(name='Plot all Polar plots and save to file', button_type='primary')

    def __init__(self, filename, client_input):
        super().__init__()

        self.ignore_overall = False
        self.client = client_input
        self.filename = Path(filename)
        self._update_dataset()

    def _update_dataset(self):
        self.ds = xr.open_dataset(self.filename,
                                  chunks={'Orientation': 1, 'wavelength': 14, 'x': -1, 'y': -1, 'Polarization': -1},
                                  engine="zarr")  # chunked for heatmap selected
        try:
            fit_ver = self.ds.attrs["fit_version"]
        except:
            fit_ver = 0
        try:
            data_ver = self.ds.attrs["data_version"]
        except:
            data_ver = 0
        current_fit_version = 1
        current_data_version = 2
        if fit_ver < current_fit_version:
            time1 = time.time()
            self.ds = self.ds.drop_vars(["fitted", "covariance"], errors="ignore")
            self.ds["navigation"] = self.ds["ds1"].mean(dim='Polarization').compute()  # chunked for navigation
            self.ds["heatmap_all"] = self.ds["ds1"].mean(dim=['x', 'y']).compute()  # chunked for heatmap all
            self.ds = self.ds.merge(self.ds["heatmap_all"].curvefit(["Polarization"], function,
                                                                    param_names=["delta", "A", "B", "theta", "C"]
                                                                    , kwargs={"maxfev": 1000000, "xtol": 10 ** -9,
                                                                              "ftol": 10 ** -9}).compute())
            self.ds.attrs["fit_version"] = current_fit_version
            self.ds = self.ds.rename({"curvefit_coefficients": "fitted", "curvefit_covariance": "covariance"})
            self.ds.to_zarr(self.filename, mode="a", compute=True)
            time2 = time.time()
            print(f"finished in {str(timedelta(seconds=time2 - time1))}")
        if data_ver == 1:
            print("fixing dataset innaccuracies")
            self.ds.attrs["data_version"] = current_data_version
            self.ds.coords['Polarization'] = np.arange(0, 180, 1) / 90 * np.pi
            self.ds.coords['degrees'] = ("Polarization", np.arange(0, 360, 2))
            self.ds.to_zarr(self.filename, mode="a", compute=True)
        self.coords = self.ds["ds1"].coords
        self.attrs = {**self.ds["ds1"].attrs, **self.ds.attrs}
        self.fname = self.attrs["title"]
        self.x = hv.Dimension('x', unit=self.attrs['x'])
        self.y = hv.Dimension('y', unit=self.attrs['y'])
        self.param['wavelength'].objects = self.ds["ds1"].coords['wavelength'].values.tolist()
        self.param['power'].objects = self.ds["ds1"].coords['power'].values.tolist()
        self.wavelength = self.ds["ds1"].coords["wavelength"].min().values
        self.power = self.ds["ds1"].coords["power"].min().values
        self.polarization_dim = hv.Dimension('Polarization', range=utils.get_range('Polarization', self.coords),
                                             unit=self.attrs['Polarization'])
        self.wavelength_dim = hv.Dimension('wavelength', range=utils.get_range('wavelength', self.coords),
                                           unit=self.attrs['wavelength'])

    def fit(self):
        """
        Fits the currently selected data. Only used when selected, otherwise fitted at load time and saved to file
        :return: curve fit results
        """
        if self.selected:
            output = self.ds["ds1"].sel(Orientation=self.orientation, wavelength=self.wavelength, power=self.power).sel(
                x=slice(self.x0, self.x1), y=slice(self.y0, self.y1))
        else:
            output = self.ds["ds1"].sel(Orientation=self.orientation, wavelength=self.wavelength, power=self.power)
        curvefit = output.curvefit(["Polarization"], function, reduce_dims=["x", "y"],
                                   param_names=["delta", "A", "B", "theta", "C"])
        curvefit_coefficients = curvefit.curvefit_coefficients  # idk what to do with the covars
        return curvefit_coefficients.values

    @param.depends('orientation', 'wavelength', 'colorMap', 'power')
    def nav(self):
        """
        renders plot for navigation: XY for specified Orientation, wavelength, and power, averaged against polarization
        :return:
        """
        polys = self.poly_generate()  # copy existing selection
        box_stream = streams.BoxEdit(source=polys, num_objects=1)  # create BoxEdit stream
        box_stream.add_subscriber(self.tracker)  # adds the subscriber to the BoxEdit stream
        output = self.ds["navigation"].sel(Orientation=self.orientation, wavelength=self.wavelength, power=self.power)
        self.zdim = hv.Dimension('Intensity', range=(output.values.min(), output.values.max()))
        opts = [hv.opts.Image(colorbar=True, height=600,
                              width=round(600 * self.ds["ds1"].coords['x'].size / self.ds["ds1"].coords['y'].size),
                              title=f"Wavelength: {self.wavelength}, Orientation: {self.orientation}",
                              cmap=self.colorMap, tools=['hover'], framewise=True, logz=True)]
        return hv.Image(output, vdims=self.zdim).opts(opts).redim(x=self.x, y=self.y) * polys

    def poly_generate(self):
        """
        regenerates existing polygons if they exist. This allows selections to persist across wavelengths/orientations/powers
        :return:
        """
        if self.selected:
            avg_x = (self.x0 + self.x1) / 2
            avg_y = (self.y0 + self.y1) / 2
            width = self.x1 - self.x0
            height = self.y1 - self.y0
            return hv.Polygons([hv.Box(avg_x, avg_y, (width, height))]).opts(fill_alpha=0.2, line_color='white')
        return hv.Polygons([]).opts(fill_alpha=0.2, line_color='white')

    def tracker(self, data):
        """
        Takes data from BoxEdit stream and updates variables accordingly

        :param data: data, supplied by boxedit stream
        :return:
        """
        if not data or not any(len(d) for d in data.values()):
            self.selected = False
        else:
            self.x0 = round(data['x0'][0], 3)
            self.x1 = round(data['x1'][0], 3)
            self.y0 = round(data['y0'][0], 3)
            self.y1 = round(data['y1'][0], 3)
            self.selected = True

    @param.depends('orientation', 'wavelength', 'colorMap', 'x1', 'x0', 'y0', 'y1', 'selected', 'power')
    def title(self):
        """
        renders the title
        :return:
        """
        if self.selected:
            return pn.pane.Markdown(
                f'''##{self.fname}: Orientation: {self.orientation}, wavelength: {self.wavelength}, power: {self.power},
                x0: {self.x0},x1: {self.x1}, y0: {self.y0}, y1: {self.y1}''',
                width=1800)
        return pn.pane.Markdown(
            f'''##{self.fname}: Orientation: {self.orientation},wavelength: {self.wavelength}, power: {self.power}, 
                Average across all points''',
            width=1800)

    @param.depends('orientation', 'wavelength', 'colorMap', 'x1', 'x0', 'y0', 'y1', 'selected', 'power')
    def heat_map(self):
        """
        generates the heatmap Polarization x Wavelength for given orientation
        :return:
        """
        if not self.selected:
            output = self.ds["heatmap_all"].sel(Orientation=self.orientation, power=self.power)
            title = f'''{self.fname}: Orientation: {self.orientation}, Average across all points'''
        else:
            output = self.ds["ds1"].sel(Orientation=self.orientation, power=self.power).sel(
                x=slice(self.x0, self.x1),
                y=slice(self.y0,
                        self.y1)).mean(
                dim=['x', 'y'])
            title = f'''{self.fname}: Orientation: {self.orientation}, x0: {self.x0},x1: {self.x1}, y0: {self.y0}, y1: 
                    {self.y1}'''
        line = hv.HLine(self.wavelength).opts(line_width=600 / self.coords['wavelength'].values.size, alpha=0.6)
        opts = [
            hv.opts.Image(cmap=self.colorMap, height=600, width=900, colorbar=True, title=title, tools=['hover'],
                          framewise=True, logz=True)]
        return hv.Image(output).opts(opts).redim(wavelength=self.wavelength_dim,
                                                 Polarization=self.polarization_dim) * line

    @param.depends('orientation', 'wavelength', 'x1', 'x0', 'y0', 'y1', 'selected', 'power')
    def polar(self, dataset="Polarization"):
        """
        Generates the polar plot. If self.selected, then filter to x1,x0,y0,y1
        :param dataset: not entirely sure why this is here, but this supplies the coordinate for theta_values.
        :return:
        """
        theta_values = self.coords[dataset].values  # units, plotly, radians and degrees are a mess.
        theta_radians = self.coords['Polarization'].values
        if self.selected:
            output = self.ds["ds1"].sel(Orientation=self.orientation, wavelength=self.wavelength, power=self.power,
                                        x=slice(self.x0, self.x1), y=slice(self.y0, self.y1)).mean(dim=['x', 'y'])
            data_frame = pd.DataFrame(
                np.vstack((output, theta_values, np.tile("Raw Data, over selected region", 180))).T,
                columns=['Intensity', 'Polarization', 'Data'], index=theta_values)
            fitted = function(theta_radians, *self.fit())
            data_frame2 = pd.DataFrame(
                np.vstack((fitted, theta_values, np.tile("Fitted Data, to selected points", 180))).T,
                columns=['Intensity', 'Polarization', 'Data'], index=theta_values)
            data_frame = data_frame.append(data_frame2)
        else:
            overall = self.ds["heatmap_all"].sel(Orientation=self.orientation, wavelength=self.wavelength,
                                                 power=self.power)
            data_frame = pd.DataFrame(np.vstack((overall, theta_values, np.tile("Raw Data, over all points", 180))).T,
                                      columns=['Intensity', 'Polarization', 'Data'], index=theta_values)
            params = self.ds["fitted"].sel(Orientation=self.orientation, wavelength=self.wavelength,
                                           power=self.power).values
            fitted = function(theta_radians, *params)
            data_frame2 = pd.DataFrame(np.vstack((fitted, theta_values, np.tile("Fitted Data, to all points", 180))).T,
                                       columns=['Intensity', 'Polarization', 'Data'], index=theta_values)
            data_frame = data_frame.append(data_frame2)
        data_frame = data_frame.astype({'Polarization': 'float', 'Intensity': "float", "Data": "string"})
        return px.scatter_polar(data_frame, theta="Polarization", r='Intensity', color='Data',
                                start_angle=0,
                                direction="counterclockwise",
                                range_r=(data_frame['Intensity'].min() * 0.8, data_frame['Intensity'].max() * 1.2), )

    def polars_to_file(self, event=None):
        """
        Plots all the polars to a file
        :param event: required for the button
        :return:
        """
        start = time.time()
        ori_wavelength, ori_orientation, ori_power = self.wavelength, self.orientation, self.power
        self.param["Orientation"].precedence = -1
        self.param["wavelength"].precedence = -1
        self.param["power"].precedence = -1
        wavelengths = self.ds["ds1"].coords['wavelength'].values.tolist()
        orientations = self.ds["ds1"].coords['Orientation'].values.tolist()
        powers = self.ds["ds1"].coords['power'].values.tolist()
        folder = str(self.filename).replace(f".{utils.extension(self.filename)}", '')
        if not os.path.isdir(folder):
            os.mkdir(folder)
        for power in powers:
            self.power = power
            for orientation in orientations:
                selected = self.selected
                self.orientation = orientation
                self.selected = selected
                for wavelength in wavelengths:
                    selected = self.selected
                    self.wavelength = wavelength
                    self.selected = selected  # prevent it resetting like it should
                    if self.selected:
                        title = f"{folder}/Polar_X{self.x0}:{self.x1}Y{self.y0}:{self.y1},O{orientation}W{wavelength}" \
                                f"P{power}.png "
                    else:
                        title = f"{folder}/Polar_O{orientation}W{wavelength}.png"
                    self.polar("degrees").write_image(title)
        self.nav()
        self.param["Orientation"].precedence = 1
        self.param["wavelength"].precedence = 1
        self.param["power"].precedence = 1
        self.wavelength, self.orientation, self.power = ori_wavelength, ori_orientation, self.power
        end = time.time()
        print(end - start)

    def xarray(self) -> pn.panel:
        """
        Renders the xarray dataset as a panel
        :return:
        """
        return pn.panel(self.ds, width=700)

    def view(self) -> pn.Column:
        """
        Renders everything but the widgets as a view
        :return: view of graphs and title
        """
        return pn.Column(self.title, pn.Row(self.nav, self.heat_map), pn.Row(self.polar, self.xarray))

    def widgets(self) -> pn.Column:
        """
        Renders the widgets
        :return: the widgets
        """
        self.button.on_click(self.polars_to_file)
        if self.ds["ds1"].coords["power"].size > 1:
            widgets = {"wavelength": pn.widgets.DiscreteSlider, "power": pn.widgets.DiscreteSlider}
        else:
            widgets = {"wavelength": pn.widgets.DiscreteSlider}
            self.param["power"].precedence = -1  # effectively a 5d graph
        return pn.Column(pn.Param(self.param, widgets=widgets), self.button)



import time

import holoviews as hv
import neogiinstruments
import numpy as np
import panel as pn
import param

from ..ensemblebase import EnsembleBase, Coordinate, Coordinates
from ... import utils

name = "RASHG"

hv.extension('bokeh')


def get_calibs() -> list:
    """
    Scans for calibration files

    :return: list of all calibration files
    :rtype: list of PosixPath
    """
    return list(utils.scan_directory({"WavelengthPoweredCalib": None}).keys())


class Ensemble(EnsembleBase):
    x1 = param.Integer(default=0, bounds=(0, 2047))
    x2 = param.Integer(default=100, bounds=(0, 2047))
    y1 = param.Integer(default=0, bounds=(0, 2047))
    y2 = param.Integer(default=100, bounds=(0, 2047))
    wavstart = param.Integer(default=780)
    wavend = param.Integer(default=800)
    wavstep = param.Integer(default=2)
    pol_step = param.Number(default=2, bounds=(1, 2))  # TODO: figure out how this works
    pow_start = param.Integer(default=0)
    pow_stop = param.Integer(default=5)
    pow_step = param.Integer(default=5)
    xbin = param.Integer(default=1)
    ybin = param.Integer(default=1)  # TODO add bounds
    exp_time = param.Number(default=10000)
    escape_delay = param.Integer(default=120)  # should beep at 45
    wavwait = param.Number(default=5)
    debug = param.Boolean(default=False)
    live = param.Boolean(default=True)
    colorMap = param.ObjectSelector(default="fire", objects=hv.plotting.util.list_cmaps())
    cam = neogiinstruments.camera("Camera")
    rbot, rtop, atten = [neogiinstruments.rotator(name) for name in ["rbot", "rtop", "atten"]]
    MaiTai = neogiinstruments.MaiTai("MaiTai")
    type = name
    data = "RASHG"
    dimensions = ["wavelength", "power", "Orientation", "Polarization", "x", "y"]
    cap_coords = ["x", "y"]
    loop_coords = ["wavelength", "power", "Orientation", "Polarization"]
    calibration_file = param.ObjectSelector()

    @param.depends("debug", watch=True)
    def no_wait(self):
        if self.debug:
            self.wavwait = 0
            self.escape_delay = 0
            self.exp_time = 0

    def start(self):
        print("Gathering Data, Get Out")
        self.rbot.instrument.home()
        self.rtop.instrument.home()
        self.atten.instrument.home()
        if not self.debug:
            # time.sleep(120)
            pass

    def __init__(self):
        files = get_calibs()
        if len(files) == 0:
            print("Needs calibration file ")
        self.param["calibration_file"].objects = files
        self.param["calibration_file"].default = files[0]
        super().__init__()
        self.xDim = hv.Dimension('x', unit="micrometers")
        self.yDim = hv.Dimension('y', unit="micrometers")

    def initialize(self):
        self.initialized = True
        exclude = []
        for param in self.param:
            if not param in exclude:
                self.param[param].constant = True

        self.cam.instrument.roi(self.x1, self.x2, self.y1, self.y2)
        self.cam.instrument.binning(self.xbin, self.ybin)
        if self.xbin != self.ybin:
            print('X-bin and Y-bin must be equal, probably')
        self.init_vars()
        self.coords = Coordinates(
            [
                Coordinate("wavelength", "nanometer", "wavelength", self.wavelength, self.wav_step),
                Coordinate("power", "milliwatts", "power", self.pwr, self.pow_step_func),
                Coordinate("degrees", "degrees", "Polarization", self.Polarization),
                Coordinate("Polarization", "pixels", "Polarization", self.Polarization_radians),
                Coordinate("x_pxls", "nanometer", "x", self.x_coords),
                Coordinate("x", "micrometers", "x", self.x_mm),
                Coordinate("y_pxls", "pixels", "y", self.y_coords),
                Coordinate("y", "micrometers", "y", self.y_mm),
                Coordinate("Orientation", "?", "Orientation", self.Orientation)
            ]
        )
        # self.PC, self.PCcov, self.WavPowAng, self.pc = PCFit(self.calibration_file)

    def init_vars(self):
        self.x = self.x2 - self.x1
        self.y = self.y2 - self.y1  # TODO: fix binning
        self.wavelength = np.arange(self.wavstart, self.wavend, self.wavstep, dtype=np.uint16)
        self.pwr = np.arange(self.pow_start, self.pow_stop, self.pow_step, dtype=np.uint16)
        x = int((self.x2 - self.x1) / self.xbin)
        y = int((self.y2 - self.y1) / self.ybin)
        self.cache = self.live_call()
        self.x_coords = np.arange(x, dtype=np.uint16)
        self.x_mm = np.arange(x, dtype=np.uint16) * 0.05338  # magic
        self.y_coords = np.arange(y, dtype=np.uint16)
        self.y_mm = np.arange(y, dtype=np.uint16) * 0.05338  # magic
        self.Orientation = np.arange(0, 2)
        self.Polarization = np.arange(0, 360, self.pol_step, dtype=np.uint16)
        self.Polarization_radians = np.arange(0, 360, self.pol_step, dtype=np.uint16) * np.pi / 180
        self.pwr = np.arange(self.pow_start, self.pow_stop, self.pow_step, dtype=np.uint16)
        self.pc_reverse = utils.interpolate(self.calibration_file, self.pwr)

    def get_frame(self, coords):
        o = coords[2]
        p = coords[3]*180/np.pi
        if o == 1:
            sys_offset = 45
        else:
            sys_offset = 0
        pos = p * 180 / np.pi
        pos_top = int(pos + sys_offset)
        pos_bot = int(pos)
        if self.debug:
            print(f"Moving A to {pos_top}")
        self.rtop.instrument.move_abs(pos_top)
        if self.debug:
            print(f"Moving B to {pos_bot}")
        self.rbot.instrument.move_abs(pos_bot)
        if self.debug:
            print(f"Capturing frame")
        self.cache = self.live_call()
        return {"ds1": self.cache}

    def pow_step_func(self, xs):
        pw = xs[1]
        w = xs[0]
        atten_pos = self.pc_reverse.sel(wavelength=w, power=pw)  # technically can interp here again but don't need to
        if -360 < atten_pos < 360:
            self.atten.instrument.move_abs(atten_pos)

    def graph(self, live=False):
        if live:
            self.cache = self.live_call()
        output = self.cache
        self.zdim = hv.Dimension('Intensity', range=(output.min(), output.max()))
        opts = [hv.opts.Image(colorbar=True, cmap=self.colorMap, tools=['hover'], framewise=True, logz=True)]
        return hv.Image(output, vdims=self.zdim).opts(opts).redim(x=self.xDim, y=self.yDim)

    def live_call(self):
        return self.cam.instrument.get_frame(exp_time=self.exp_time)

    def wav_step(self, xs):
        time.sleep(self.wavwait)
        self.MaiTai.instrument.Set_Wavelength(xs[0])
        self.pow_step_func(xs)

    def widgets(self):
        if self.initialized:
            return pn.Column(self.atten.view, self.rbot.view, self.rtop.view, self.cam.view, self.MaiTai.view)
        else:
            return None



"""
Contains Viewer class
"""
from typing import Union

from dask.distributed import Client
import param
import panel as pn

from .. import utils


class Viewer(param.Parameterized):
    """
    Allows user to select file and loads appropriate visualizer for file
    """
    filename: str = param.ObjectSelector()

    def __init__(self, types, filename: str = None):
        self.types = types
        """        
        :param filename: specify default filename, otherwise choose files
        :type filename: str
        :param types: dictionary matching datatypes to modules to open them
        :type types: dict
        :rtype: Viewer
        """
        super().__init__()
        self.client = Client()
        self.reload_files()
        if filename is None:
            self.filename = self.files[0]
        else:
            self.filename = filename
        self.load()

    def reload_files(self) -> None:
        """
        Checks for files with given extensions

        :rtype: None

        """
        self.file_dict = utils.scan_directory(self.types)
        if len(self.file_dict) == 0:
            raise Exception("must have at least one file")
        self.files = list(self.file_dict.keys())
        self.param["filename"].objects = self.files

    @param.depends('filename', watch=True)
    def load(self) -> None:
        """
        Loads currently selected file
        """
        self.reload_files()  # temp solution
        visualizer = self.file_dict[self.filename].Grapher
        self.grapher = visualizer(self.filename, self.client)

    @param.depends('filename')
    def widgets(self) -> pn.Column:
        """
        Renders widgets for selected visualizer
        :return: pn.Column
        """
        return pn.Column(self.param, self.grapher.widgets(), self.dask)

    @param.depends('filename')
    def graph(self) -> Union[pn.Row, pn.Column]:
        """
        Returns graph from the visualizer
        :return:
        """
        return self.grapher.view()

    def dask(self):
        """
        Displays dask client
        :return:
        """
        return self.client

    def view(self) -> Union[pn.Row, pn.Column]:
        """
        Combines the widgets and graph to form a view
        :return: view
        :rtype: Union[pn.Row, pn.Column]
        """
        return pn.Row(self.widgets, self.graph)

    def stop(self):
        """
        Tries to stop the visualizer
        Currently doesn't do anything
        :return:
        """
        pass



"""
Provides collections of ensembles for gathering data
"""
import os
from pathlib import Path
from importlib import import_module
from . import gui
instruments = {}
for directory in os.scandir(str(Path(__file__).resolve().parent)):
    if directory.is_dir():
        for file in os.listdir(directory):
            if str(file).__contains__("ensemble_"):
                if not str(file).__contains__("base"):
                    sname = file.replace(".py", "")
                    try:
                        module = import_module(f"{__name__}.{directory.name}.{sname}")
                    except Exception as ex:
                        print(f"{file} import failed. Exception {Exception}")
                    else:
                        instruments[module.name] = module



from pathlib import Path

import holoviews as hv
import panel as pn
import param
import xarray as xr

from neogidashboard import utils

hv.extension("bokeh")
DATA_TYPE = "stellarnet"


class Grapher(param.Parameterized):
    wavelength = param.Selector()

    def _update_dataset(self):
        """
        Opens selected dataset
        """
        self.data = xr.open_zarr(self.filename)
        self.param['wavelength'].objects = self.data["Stellarnet"].coords['wavelength'].values.tolist()
        self.wavelength = self.data["Stellarnet"].coords["wavelength"].min().values
        self.power_dim = hv.Dimension("power", range=utils.get_range("power", self.data["Stellarnet"].coords))

    def __init__(self, filename, client_input):
        """
        :param filename: filename to open
        :param client_input: Dask Client
        """
        super().__init__()
        self.client = client_input
        self.filename = Path(filename)
        self._update_dataset()

    @param.depends("wavelength")
    def maximum(self):
        """
        Maximum graph of intensity for all emission wavelength against power per wavelength
        :return:
        """

        output = self.data["Stellarnet"].sel(wavelength=self.wavelength).idxmax("emission_wavelength")
        opts = [hv.opts.Curve(title=f"Wavelength: {self.wavelength}",
                              tools=['hover'], framewise=True)]
        return hv.Curve(output, "power", "emission wavelength for maximum intensity").opts(opts)

    @param.depends("wavelength")
    def integrated(self):
        output = self.data["Stellarnet"].sel(wavelength=self.wavelength).sum(dim="emission_wavelength")
        opts = [hv.opts.Curve(title=f"Wavelength: {self.wavelength}",
                              tools=['hover'], framewise=True)]
        return hv.Curve(output, self.power_dim, "summed intensity").opts(opts)

    @param.depends("wavelength")
    def overall(self):
        output = self.data["Stellarnet"].sel(wavelength=self.wavelength)
        opts = [hv.opts.QuadMesh(colorbar=True, tools=['hover'], framewise=True, logz=True, width=600)]
        return hv.QuadMesh(output, [self.power_dim, "emission_wavelength"]).opts(opts)

    def view(self):
        """
        wraps nav
        :return: navigation
        """
        return pn.Column(pn.Row(self.integrated, self.maximum), self.overall)

    def widgets(self):
        """
        returns widgets
        :return: widgets
        """
        widgets = {"wavelength": pn.widgets.DiscreteSlider}
        return pn.Param(self.param, widgets=widgets)


if __name__ == "__main__":
    viewer = Grapher("data/realtest.zarr", None)
    viewer.view().show()
    # print(viewer.power_dim)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['neogidashboard',
 'neogidashboard.ensembles',
 'neogidashboard.ensembles.rashg',
 'neogidashboard.ensembles.stellarnet',
 'neogidashboard.visualizer',
 'neogidashboard.visualizer.rashg',
 'neogidashboard.visualizer.stellarnet']

package_data = \
{'': ['*']}

install_requires = \
['bokeh>=2.3.2,<3.0.0',
 'colorcet>=2.0.6,<3.0.0',
 'dask[distributed]>=2021.7.0,<2022.0.0',
 'hvplot>=0.7.2,<0.8.0',
 'matplotlib>=3.4.2,<4.0.0',
 'neogiinstruments>=2.6.0,<3.0.0',
 'netCDF4>=1.5.7,<2.0.0',
 'numba>=0.53.1,<0.54.0',
 'numpy>=1.21.0,<2.0.0',
 'pandas>=1.2.5,<2.0.0',
 'panel>=0.11.3,<0.12.0',
 'plotly>=5.0.0,<6.0.0',
 'scipy>=1.7.0,<2.0.0',
 'simple-pid>=1.0.1,<2.0.0',
 'typing-extensions>=3.10.0,<4.0.0',
 'xarray>=0.18.2,<0.19.0',
 'zarr>=2.3,<3.0']

entry_points = \
{'console_scripts': ['clean = dashboard.console_utilities:clean',
                     'converter = neogidashboard.console_utilities:convert',
                     'dashboard = neogidashboard.combined:main']}

setup_kwargs = {
    'name': 'neogidashboard',
    'version': '0.1.1',
    'description': 'Dashboard for visualizing data',
    'long_description': None,
    'author': 'bageljr',
    'author_email': 'aryanagarwal897@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9.6,<3.10.0',
}


setup(**setup_kwargs)



"""
houses gui class
"""
import os
import sys
import time

import numpy as np
import panel as pn
import param
import xarray as xr
import zarr
from numba import njit
from tqdm.contrib.itertools import product

from .ensemblebase import EnsembleBase, Coordinates

compressor = zarr.Blosc(cname="zstd", clevel=3, shuffle=2)


@njit(cache=True)
def compare(coords, dim_cache):
    """
    Get first different dimension
    :param coords: new coordinates
    :param dim_cache: old coordinates
    :return: index of changed coordinate
    """
    for i, coord in enumerate(coords):
        if coord > dim_cache[i]:
            return i
    raise Exception("Coordinates identical")


class Gui(param.Parameterized):
    """
    Essentially the backbone of the data collection
    Loops through dimensions
    """
    ensemble: EnsembleBase
    coordinates: Coordinates
    c_pol = param.Number(default=0, precedence=-1)
    institution = param.String(default="University of North Texas")
    sample = param.String(default="MoS2")
    GUIupdate = param.Boolean(default=True)
    button = pn.widgets.Button(name='Gather Data', button_type='primary')
    live = param.Boolean(default=False, precedence=-1)
    refresh = 5  # refresh every 5 seconds #make it a parameter
    live_refresh = param.Integer(default=5)
    dim_cache = np.array([0, 0, 0, 0])
    ensembles = param.ObjectSelector()  # Initializes a blank object selector, fills it in later
    confirmed = param.Boolean(default=False, precedence=-1)
    button2 = pn.widgets.Button(name='Confirm', button_type='primary')
    cap_coords: dict
    attrs: dict
    datasets: dict
    mask: dict
    encoders: dict
    data: xr.Dataset

    def __init__(self, ensembles):
        self.ensemble_classes = ensembles
        ensembles = list(ensembles.keys())
        i = 0
        while i < len(ensembles):
            try:
                self.param["ensembles"].default = ensembles[i]
            except:
                print(f"{ensembles[i]} failed")
            i += 1
        self.param["ensembles"].objects = ensembles
        super().__init__()
        self.callback = pn.state.add_periodic_callback(self.live_view, period=self.live_refresh * 1000, start=False)
        self.button.disabled = True
        self.load()
        self.button2.on_click(self.initialize)

    @param.depends('ensembles', watch=True)
    def load(self) -> None:
        """
        Loads selected ensemble
        :rtype: None
        """
        self.confirmed = False
        try:
            self.ensemble = self.ensemble_classes[self.ensembles].Ensemble()
        except:
            print("Ensemble didn't initialize")

    def initialize(self, event=None):
        """
        Initializes the GUI. Sets all parameters to constant and allows the user to start the expirement. If the live
        view is enabled, start the callback.

        :return: None
        """
        self.ensemble.initialize()
        self.init_vars()
        self.button2.disabled = True
        self.confirmed = True
        if self.ensemble.gather:
            self.button.disabled = False
        self.button.on_click(self.gather_data)
        if self.ensemble.live:
            self.live = True
            self.callback.start()
        exclude = ["c_pol", "live"]
        for parameter in self.param:
            if parameter not in exclude:
                self.param[parameter].constant = True

    def init_vars(self):
        """
        Initialize all the variables. Should probably be in initialize
        :return:
        """
        # populate metadata
        self.attrs = {
            "title": self.ensemble.title,
            "institution": self.institution,
            "sample": self.sample,
            "source": self.ensemble.type,
            "data_type": self.ensemble.data,
            "time": time.strftime('%a, %d %b %Y %H:%M:%S', time.localtime()),
            "fit_version": 0,
            "data_version": 2
        }
        self.coordinates = self.ensemble.coords
        self.coords = {coord.name: ([coord.dimension], coord.values) for coord in self.coordinates}
        for coord in self.coordinates:
            self.attrs[coord.name] = coord.unit

        # data.date = str(datetime.date.today()) #out of date
        # create variables; in this case, the only dependent variable is 'shg',
        # which is the shg intensity along the specified dimensions
        # populate coordinate dimensions
        dimensions = self.ensemble.dimensions
        if type(dimensions) is list:
            dimensions = {dataset: dimensions for dataset in self.ensemble.datasets}
        self.cap_coords = self.ensemble.cap_coords
        if type(self.cap_coords) is list:
            self.cap_coords = {dataset: self.cap_coords for dataset in self.ensemble.datasets}
        zeros = {}
        for dataset in self.ensemble.datasets:
            zero_array = [self.ensemble.coords[coord].values.size for coord in dimensions[dataset]]
            zero_array[0] = 1
            zeros[dataset] = xr.DataArray(np.zeros(zero_array), dims=dimensions[dataset])
        self.datasets = {dataset: (dimensions[dataset], zeros[dataset], self.attrs) for dataset in
                         self.ensemble.datasets}
        self.encoders = {dataset: {"compressor": compressor} for dataset in self.ensemble.datasets}
        # Get filename
        fname = self.ensemble.filename
        i = 2
        while os.path.isdir(self.ensemble.filename):
            self.ensemble.param["filename"].constant = False
            self.ensemble.filename = fname.replace(".zarr", f"{i}.zarr")
            i += 1
            print(f"Zarr store exists, trying {self.ensemble.filename}")
            self.ensemble.param["filename"].constant = True
        os.makedirs(self.ensemble.filename)

    def gather_data(self, event=None):
        """
        Starts data gathering loop. Iterates over self.instruments.loop_coords
        :param event: needed for button
        :return:
        """
        if self.ensemble.live:
            self.callback.stop()
        self.button.disabled = True
        self.button2.disabled = True
        self.live = False
        first = 0
        ranges = [self.ensemble.coords[coord].values for coord in self.ensemble.loop_coords]
        self.ensemble.start()
        '''
        The infinite loop:
        not because it is an actual infinite loop but because it supports a theoretical infinite number of dimensions
        memory limits nonwithstanding
        The generator should be lazy and not overflow your memory. Theoretically.
        '''
        try:
            self.mask = {dim: min(self.ensemble.coords[dim].values) for dim in self.ensemble.loop_coords}
        except ValueError:
            print("Empty list or 0 pol step. Check your parameters")
            sys.exit()
        for coords in product(*ranges):
            dim, dim_num = self.find_dim(coords)
            if dim_num == 0:
                if first == 0:
                    first = 1
                elif first == 1:
                    self.data.to_zarr(self.ensemble.filename, encoding=self.encoders)
                    first += 1
                else:
                    self.data.to_zarr(self.ensemble.filename, append_dim=self.ensemble.loop_coords[0])
                self.coords[self.ensemble.loop_coords[0]] = ([dim], [coords[0]])  # update 1st coord after saving
                self.data = xr.Dataset(
                    data_vars=self.datasets,
                    coords=self.coords,
                    attrs=self.attrs
                )
            self.mask[dim] = coords[dim_num]
            self.ensemble.coords[dim](coords)
            data = self.ensemble.get_frame(coords)
            for dataset in self.ensemble.datasets:
                self.data[dataset].loc[self.mask] = xr.DataArray(data[dataset],
                                                                 dims=self.cap_coords[dataset])
            if self.GUIupdate:
                self.c_pol += 1  # refresh the GUI
        self.data.to_zarr(self.ensemble.filename, append_dim=self.ensemble.loop_coords[0])
        self.ensemble.stop()
        print("Finished")
        self.c_pol = self.c_pol + 1
        self.data.close()
        sys.exit()

    def find_dim(self, xs):
        """
        finds the changed dimension and the index of the changed dimension
        :param xs:
        :return: changed dimension name and index of changed dimension
        """
        typed_a = np.array(xs, dtype=np.float64)
        dim = compare(typed_a, self.dim_cache)
        self.dim_cache = typed_a
        return self.ensemble.loop_coords[dim], dim

    def live_view(self):
        """
        Actually all this does is update the c_pol parameter to make param update the live view
        :return:
        """
        if self.live:
            if self.ensemble.debug:
                print("Updating live view")
            self.c_pol = self.c_pol + 1

    @param.depends('c_pol')
    def graph(self) -> pn.Row():
        """
        returns the graph from the ensemble

        :return:
        """
        if self.confirmed and self.ensemble.live:
            return self.ensemble.graph(live=self.live)
        return pn.Row()

    def widgets(self) -> pn.Row:
        """
        Renders everything but the graph
        :return:
        """
        return pn.Row(pn.Column(self.param, self.button2, self.button), self.ensemble.param,
                      self.ensemble.widgets)

    def stop(self):
        """
        Stops the server
        :return:
        """
        if self.ensemble.live:
            self.callback.stop()
        print("shutting down live view")  # doesn't currently  work



"""
Visualizer for various data types
"""
import os
from pathlib import Path
from importlib import import_module
from .visualizer import Viewer

types = {}
for directory in os.scandir(str(Path(__file__).resolve().parent)):
    if directory.is_dir():
        for file in os.listdir(directory):
            if str(file).__contains__("visualizer_"):
                sname = file.replace(".py", "")
                try:
                    module = import_module(f"{__name__}.{directory.name}.{sname}")
                except ImportError:
                    print(f"{file} import failed")
                else:
                    types[module.DATA_TYPE] = module
